# -*- python -*-
#
# Copyright 2021, 2022, 2023 Cecelia Chen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# djfim.contrib.diff

from ..runtime import DynamicBlock


class SDIFF(object):
    SEP = '|'
    COL_NAMES = ['status', 'choice', 'uri', 'a', 'b']

    LABEL_ADD = 'A'
    LABEL_CHANGE = 'C'
    LABEL_DROP = 'D'

    def __str__(self):
        return self.SEP.join([ str(getattr(self, f,  '')) for f in self.COL_NAMES ])

    @property
    def type(self):
        if self.a is None:
            return self.LABEL_ADD
        if self.b is None:
            return self.LABEL_DROP
        return self.LABEL_CHANGE


class DIFFBATCH(object):
    def __init__(self):
        super().__init__()
        self.load_preset()

    def load_reset(self):
        self._batch = list()
        return self

    def append(self, item):
        return self._batch.append(item)

    def get(self, uri):
        obj = None
        for item in self._batch:
            if item.uri == uri:
                obj = item
                break
        return obj

    def getEntityStatus(self, uri):
        diff = self.get(uri)
        return diff.status

    def getEntityChoice(self, uri):
        diff = self.get(uri)
        return diff.choice


class DiffLoader(DynamicBlock):
    CLASS_DIFF = 'djfim.contrib.diff.SDIFF'
    CLASS_BATCH = 'djfim.contrib.diff.DIFFBATCH'

    def __init__(self):
        super().__init__()
        self.load_preset()

    def load_preset(self):
        self._diff_kls = self.getDynClass(self.CLASS_DIFF)
        self._batch_kls = self.getDynClass(self.CLASS_BATCH)
        return self

    def loadFromString(self, content):
        cache = self._batch_kls()
        for row in content.splitlines():
            obj = self.parse_single_row(row)
            cache.append(obj)
        return cache

    def parse_single_row(self, line):
        obj = self._diff_kls()

        values = line.split(obj.SEP)
        assert len(values) >= len(obj.COL_NAMES), 'invalid input'

        for field in obj.COL_NAMES:
            fval = values[ obj.COL_NAMES.index(field) ]
            setattr(obj, field, fval)

        return obj


class DiffScanner(DynamicBlock):
    CLASS_ENTITY_DIFF = 'djfim.contrib.diff.EntityDiff'

    CLASS_PACK_SOLVER  = 'djfim.contrib.partition.PackSolver'
    CLASS_ALIGN_SOLVER = 'djfim.contrib.partition.Alignment'

    def __init__(self, context, aHead, bHead):
        super().__init__()
        self.load_preset()
        #
        self._context = context
        self._a_head = aHead
        self._b_head = bHead

    def load_preset(self):
        self._pack_solver_cls = self.getDynClass(self.CLASS_PACK_SOLVER)
        self._entity_diff_cls = self.getDynClass(self.CLASS_ENTITY_DIFF)
        self._align_solver_cls = self.getDynClass(self.CLASS_ALIGN_SOLVER)
        self._dependency_cache_cls = dict
        return self

    def getPackPair(self):
        # a) get sorted;
        sorted_pack_name = self._context.getSortSolver().get()

        # b) collect packs;
        a_pak_solver = self._pack_solver_cls(self._a_head)
        b_pak_solver = self._pack_solver_cls(self._b_head)

        for name in  sorted_pack_name:
            a_pack = a_pak_solver.getPack(name)
            b_pack = b_pak_solver.getPack(name)
            yield (a_pack, b_pack)

    def getAlignedCollection(self, aSet, bSet):
        align = self._align_solver_cls(aSet, bSet)
        align.item_label_attr = 'anchor'
        return align.do().output

    def filterStatus(self, flag):
        rval = False
        if flag[0] in ('L', 'D'):
            rval = True
        return rval

    def filterDiff(self, item):
        rval = False
        try:
            if item.type in (item.LABEL_ADD, item.LABEL_CHANGE):
                rval = True
        except:
            pass
        return rval

    def getDiff(self, cache):
        self._dependency_cache = self._dependency_cache_cls()
        _diff_cache = list()
        # a) get pack/partition on head;
        for each_pack in self.getPackPair():
        # b) scan within each pack, register any mapping in the `_dependency_cache`;
            pack_status = self.getAlignedCollection(
                each_pack[0],
                each_pack[1]
            )
            for flag in pack_status:
                if self.filterStatus(flag):
                    diff = self._entity_diff_cls(
                        self._b_head.pool.getEntity(each_pack[0].get(flag[1])),
                        self._b_head.pool.getEntity(each_pack[1].get(flag[2])),
                        self._dependency_cache
                    )
                    _diff_cache.append(diff.get())
        # c) concatenate rvalue;
        self.diff_cache = [ item for item in _diff_cache if self.filterDiff(item) ]
        cache.extend(self.diff_cache)
        return self


class EntityDiff(DynamicBlock):
    CLASS_DIFF = 'djfim.contrib.diff.SDIFF'

    def __init__(self, aEntity, bEntity, dependency):
        super().__init__()
        self.load_preset()
        #
        self._a_entity = aEntity
        self._b_entity = bEntity
        self._dependency = dependency

    def load_preset(self):
        self._diff_cls = self.getDynClass(self.CLASS_DIFF)
        return self

    def lookUpDependency(self, anchor, f):
        rval = None
        try:
            obj = self._dependency[ f ].get(anchor, None)
            rval = obj.anchor
        except:
            pass
        return rval

    def checkDiff(self):
        rval = True
        try:
            rval = not( self._a_entity.content == self._b_entity.content )
        except:
            pass
        return rval

    def constructDiff(self):
        diff = self._diff_cls()
        diff.a = self._a_entity.content
        diff.b = self._b_entity.content
        diff.uri = self._a_entity.uri
        return diff

    def get(self):
        obj = None
        if self.checkDiff():
            obj = self.constructDiff()
        return obj


class DiffActuator(DynamicBlock):
    CLASS_DIFF = 'djfim.contrib.diff.SDIFF'

    def __init__(self, aHead, bHead, diffSet,  altPool=None):
        assert bHead is not None,  'invalid input'
        assert (aHead is not None) or (altPool is not None),  'invalid input'
        super().__init__()
        self.load_preset()
        #
        self._a_head = aHead
        self._b_head = bHead
        self._diff_set = diffSet
        self._alternative_pool = altPool

    def load_preset(self):
        self._diff_cls = self.getDynClass(self.CLASS_DIFF)
        self._dependency_cache_cls = dict
        return self

    def patchData(self, data, head):
        data['generation'] = head.storage_pk
        return data

    def applyDiff(self):
        # a) loop over existing ones;
        for entity in self._start:
            new_data = entity.copyAsDict()
            try:
                diff = self._diff_set.getDiff(entity.uri)
                if diff.type == diff.LABEL_CHANGE:
                    if diff.choice == 'b':
                        new_data = self._b_head.lookUpEntity(entity.uri).copyAsDict()
            except:
                pass

            self._pool.addEntity(
                self.patchData(
                    new_data,
                    self._new_a_head
                )
            )

        # b) collect new ones;
        for diff in self._diff_set:
            if diff.type == diff.LABEL_ADD:
                add_data = self._b_head.lookUpEntity(diff.b_uri).copyAsDict()
                self._pool.addEntity(
                    self.patchData(
                        add_data,
                        self._new_a_head
                    )
                )
        return self

    def getHeadData(self):
        d = {
            'time_stamp': self._ts,
            'memo': 'combined node',
        }
        return d

    def do(self, now):
        self._ts = now
        # a) create new node;
        new_a_data = self.getHeadData()

        try:
            self._pool = self._a_head.pool
        except AttributeError:
            self._pool = self._alternative_pool

        self._new_a_head = self._pool.addHead(
            new_a_data,
            aHead=self._a_head,
            bHead=self._b_head
        )
        # b) loop through;
        self._start = self._a_head.getEntitySet()
        self.applyDiff()
        return self

    @property
    def latest_head(self):
        return self._new_a_head
